@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: iso2022.c,v 1.24 2003/04/18 00:14:54 nalin Exp $"
+#ident "$Id: iso2022.c,v 1.28 2003/04/24 02:50:54 nalin Exp $"
 #include "../config.h"
 #include <sys/types.h>
 #include <errno.h>
@@ -52,7 +52,7 @@
 /* Fudge factor we add to wide map identifiers to keep them distinct. */
 #define WIDE_FUDGE	0x100000
 /* An invalid codepoint. */
-#define INVALID_CODEPOINT 0x3F
+#define INVALID_CODEPOINT 0xFFFF
 
 struct _vte_iso2022_map {
 	gulong from;
@@ -263,7 +263,8 @@ static const struct _vte_iso2022_map _vte_iso2022_map_wide_A[] = {
 static const struct _vte_iso2022_map _vte_iso2022_map_wide_B[] = {
 #include "unitable.JIS0208"
 };
-/* Korean.  KS X 1001, per Ken Lunde's CJKV Information Processing. */
+/* Korean.  KS X 1001 (formerly KS C 5601), per Ken Lunde's
+ * CJKV_Information_Processing. */
 static const struct _vte_iso2022_map _vte_iso2022_map_wide_C[] = {
 #include "unitable.KSX1001"
 };
@@ -378,7 +379,7 @@ _vte_iso2022_map_get(gunichar mapname,
 	if (and_mask) {
 		*and_mask = (~(0));
 	}
-	
+
 	/* Make sure we have a map, erm, map. */
 	if (maps == NULL) {
 		maps = g_tree_new(_vte_direct_compare);
@@ -511,7 +512,7 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_at,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_at));
 		}
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'A' + WIDE_FUDGE:
@@ -519,7 +520,7 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_A,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_A));
 		}
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'B' + WIDE_FUDGE:
@@ -527,7 +528,7 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_B,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_B));
 		}
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'C' + WIDE_FUDGE:
@@ -535,7 +536,7 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_C,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_C));
 		}
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'D' + WIDE_FUDGE:
@@ -543,7 +544,7 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_D,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_D));
 		}
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'G' + WIDE_FUDGE:
@@ -551,10 +552,10 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_G,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_G));
 		}
-		/* Return the plane number. */
+		/* Return the plane number as part of the "or" mask. */
 		g_assert(or_mask != NULL);
 		*or_mask = 0x10000;
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'H' + WIDE_FUDGE:
@@ -562,10 +563,10 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_G,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_G));
 		}
-		/* Return the plane number. */
+		/* Return the plane number as part of the "or" mask. */
 		g_assert(or_mask != NULL);
 		*or_mask = 0x20000;
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'I' + WIDE_FUDGE:
@@ -573,10 +574,10 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_G,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_G));
 		}
-		/* Return the plane number. */
+		/* Return the plane number as part of the "or" mask. */
 		g_assert(or_mask != NULL);
 		*or_mask = 0x30000;
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'J' + WIDE_FUDGE:
@@ -584,10 +585,10 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_G,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_G));
 		}
-		/* Return the plane number. */
+		/* Return the plane number as part of the "or" mask. */
 		g_assert(or_mask != NULL);
 		*or_mask = 0x40000;
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'K' + WIDE_FUDGE:
@@ -595,10 +596,10 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_G,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_G));
 		}
-		/* Return the plane number. */
+		/* Return the plane number as part of the "or" mask. */
 		g_assert(or_mask != NULL);
 		*or_mask = 0x50000;
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'L' + WIDE_FUDGE:
@@ -606,10 +607,10 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_G,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_G));
 		}
-		/* Return the plane number. */
+		/* Return the plane number as part of the "or" mask. */
 		g_assert(or_mask != NULL);
 		*or_mask = 0x60000;
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	case 'M' + WIDE_FUDGE:
@@ -617,10 +618,10 @@ _vte_iso2022_map_get(gunichar mapname,
 			map = _vte_iso2022_map_init(_vte_iso2022_map_wide_G,
 					    G_N_ELEMENTS(_vte_iso2022_map_wide_G));
 		}
-		/* Return the plane number. */
+		/* Return the plane number as part of the "or" mask. */
 		g_assert(or_mask != NULL);
 		*or_mask = 0x70000;
-		width = 0;
+		width = 2; /* CJKV expects 2 bytes -> 2 columns */
 		bytes = 2;
 		break;
 	default:
@@ -774,7 +775,7 @@ _vte_iso2022_fragment_input(struct _vte_buffer *input, GArray *blocks)
 {
 	unsigned char *nextctl = NULL, *p, *q;
 	const unsigned char *valids = NULL;
-	glong sequence_length;
+	glong sequence_length = 0, i;
 	struct _vte_iso2022_block block;
 	gboolean quit;
 
@@ -820,6 +821,79 @@ _vte_iso2022_fragment_input(struct _vte_buffer *input, GArray *blocks)
 				sequence_length = 0;
 			} else {
 				switch (nextctl[1]) {
+				case '[':
+					/* ESC [, the CSI.  The first letter
+					 * is the end of the sequence, */
+					for (i = 2; i < q - nextctl; i++) {
+						if (g_unichar_isalpha(nextctl[i])) {
+							break;
+						}
+						if ((nextctl[i] == '@') ||
+						    (nextctl[i] == '`') ||
+						    (nextctl[i] == '{') ||
+						    (nextctl[i] == '|')) {
+							break;
+						}
+					}
+					if (i < q - nextctl) {
+						/* Return the length of this
+						 * sequence. */
+						sequence_length = i + 1;
+					} else {
+						/* Inconclusive. */
+						sequence_length = 0;
+					}
+					break;
+				case '^':
+				case 'P':
+					/* ESC ^, the PM, or ESC P, the DCS.
+					 * Search for a string terminator. */
+					for (i = 2; i < q - nextctl - 1; i++) {
+						if ((nextctl[i] == '\033') &&
+						    (nextctl[i + 1] == '\\')) {
+							break;
+						}
+					}
+					if (i < q - nextctl - 1) {
+						/* Return the length of this
+						 * sequence. */
+						sequence_length = i + 1;
+					} else {
+						/* Inconclusive. */
+						sequence_length = 0;
+					}
+					break;
+				case ']':
+					/* ESC ], the OSC.  Search for a string
+					 * terminator or a BEL. */
+					for (i = 2; i < q - nextctl - 1; i++) {
+						if ((nextctl[i] == '\033') &&
+						    (nextctl[i + 1] == '\\')) {
+							break;
+						}
+					}
+					if (i < q - nextctl - 1) {
+						/* Return the length of this
+						 * sequence. */
+						sequence_length = i + 1;
+					} else {
+						for (i = 2;
+						     i < q - nextctl;
+						     i++) {
+							if (nextctl[i] == '\007') {
+								break;
+							}
+						}
+						if (i < q - nextctl) {
+							/* Return the length of
+							 * this sequence. */
+							sequence_length = i + 1;
+						} else {
+							/* Inconclusive. */
+							sequence_length = 0;
+						}
+					}
+					break;
 				case 'N':
 				case 'O':
 				case 'n':
@@ -896,6 +970,8 @@ _vte_iso2022_fragment_input(struct _vte_buffer *input, GArray *blocks)
 						}
 					}
 					break;
+				default:
+					break;
 				}
 			}
 			break;
@@ -1098,6 +1174,7 @@ process_control(struct _vte_iso2022_state *state, guchar *ctl, gsize length,
 		GArray *gunichars)
 {
 	gunichar c;
+	int i;
 	if (length >= 1) {
 		switch (ctl[0]) {
 		case '\r':  /* CR */
@@ -1143,6 +1220,15 @@ process_control(struct _vte_iso2022_state *state, guchar *ctl, gsize length,
 		default:
 			if ((length >= 2) && (ctl[0] == '\033')) {
 				switch (ctl[1]) {
+				case '[': /* CSI */
+				case ']': /* OSC */
+				case 'P': /* DCS */
+					for (i = 0; i < length; i++) {
+						c = (guchar) ctl[i];
+						g_array_append_val(gunichars,
+								   c);
+					}
+					break;
 				case 'N': /* SS2 */
 					state->override = 2;
 #ifdef VTE_DEBUG
@@ -1443,6 +1529,8 @@ main(int argc, char **argv)
 		{"abcd\033$(Cefgh\ri\nj\033)0k\017lmn\033Nopqrst\033%G", TRUE},
 		{"ABCD\033$(C\033)", TRUE},
 		{"0", TRUE},
+		{"\014\033[33;41m", TRUE},
+		{"\015", TRUE},
 		{"\014{|}\015\r\n", TRUE},
 		{"\033(B\033)0\033*B\033+B", TRUE},
 		{"\033$B$+$J4A;z\033(J~", TRUE},
